/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Mozilla SVG project.
 *
 * The Initial Developer of the Original Code is 
 * Crocodile Clips Ltd..
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Alex Fritze <alex.fritze@crocodile-clips.com> (original author)
 *	  James Turber <james.turner@crocodile-clips.com> (Quartz port)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

#include "nsCOMPtr.h"
#include "nsISVGQuartzCanvas.h"
#include "nsIRenderingContext.h"
#include "nsIDeviceContext.h"
#include "nsTransform2D.h"
#include "nsIPresContext.h"
#include "nsRect.h"
#include "nsSVGQuartzCanvas.h"

////////////////////////////////////////////////////////////////////////
// nsSVGGDIPlusCanvas class

class nsSVGQuartzCanvas : public nsISVGQuartzCanvas
{
public:
  nsSVGQuartzCanvas();
  virtual ~nsSVGQuartzCanvas();
  
	nsresult Init(nsIRenderingContext* ctx, 
		nsIPresContext* presContext,
        const nsRect & dirtyRect);

  // nsISupports interface:
  NS_DECL_ISUPPORTS

  // nsISVGRendererCanvas interface:
  NS_DECL_NSISVGRENDERERCANVAS  
  
  NS_IMETHOD_(CGContextRef) GetCGContext();

private:
  nsCOMPtr<nsIRenderingContext> mMozContext;
  nsCOMPtr<nsIPresContext> mPresContext;
  
  GrafPtr mGrafPtr;
  CGContextRef mCGContext;
};

//----------------------------------------------------------------------
// implementation:

nsSVGQuartzCanvas::nsSVGQuartzCanvas() :
	mGrafPtr(NULL), mCGContext(NULL)
{
 	NS_INIT_ISUPPORTS();
}

nsSVGQuartzCanvas::~nsSVGQuartzCanvas()
{
}

nsresult nsSVGQuartzCanvas::Init(nsIRenderingContext* ctx,
                         nsIPresContext* presContext,
                         const nsRect & dirtyRect)
{
	NS_ENSURE_ARG_POINTER(ctx);
	NS_ENSURE_ARG_POINTER(presContext);
	
	mPresContext = presContext;
	mMozContext = ctx;
	
	// this ctx better be what we think it is...
	/* note all we can do here is stash the graf ptr for future use :
	we can't setup the CGContext until we lock. I don't think there's any
	performance difference there though. */
	mMozContext->RetrieveCurrentNativeGraphicData((PRUint32 *)(&mGrafPtr));
	
	return NS_OK;
}

nsresult
NS_NewSVGQuartzCanvas(nsISVGRendererCanvas **result,
                       nsIRenderingContext *ctx,
                       nsIPresContext *presContext,
                       const nsRect & dirtyRect)
{
  nsSVGQuartzCanvas *canvas = new nsSVGQuartzCanvas();
  if (!canvas) return NS_ERROR_OUT_OF_MEMORY;

  NS_ADDREF(canvas);
  nsresult rv = canvas->Init(ctx, presContext, dirtyRect);

  if (NS_FAILED(rv)) {
    NS_RELEASE(canvas);
    return rv;
  }
    
  *result = canvas;
  return rv;
}

//----------------------------------------------------------------------
// nsISupports methods:

NS_IMPL_ADDREF(nsSVGQuartzCanvas)
NS_IMPL_RELEASE(nsSVGQuartzCanvas)

NS_INTERFACE_MAP_BEGIN(nsSVGQuartzCanvas)
  NS_INTERFACE_MAP_ENTRY(nsISVGRendererCanvas)
  NS_INTERFACE_MAP_ENTRY(nsISVGQuartzCanvas)
  NS_INTERFACE_MAP_ENTRY(nsISupports)
//  NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, nsISVGRendererCanvas)
NS_INTERFACE_MAP_END

//----------------------------------------------------------------------
// nsISVGRendererCanvas methods:

/* [noscript] nsIRenderingContext lockRenderingContext ([const] in nsRectRef rect); */
NS_IMETHODIMP
nsSVGQuartzCanvas::LockRenderingContext(const nsRect & rect,
                                         nsIRenderingContext **_retval)
{
// general light and magic : give unto us sweet Quartz loving
	OSErr err = ::QDBeginCGContext(mGrafPtr, &mCGContext);
	NS_ENSURE_TRUE(err == noErr, NS_ERROR_FAILURE);
	
// setup scaling / transforms	
	// We'll scale our logical 'pixles' to device units according to the
	// resolution of the device. For display devices, the canonical
	// pixel scale will be 1, for printers it will be some other value:
	nsCOMPtr<nsIDeviceContext> device;
	mMozContext->GetDeviceContext(*getter_AddRefs(device));
	float scale;
	device->GetCanonicalPixelScale(scale);
	CGContextScaleCTM(mCGContext, 1 / scale, 1 / scale);
	
	// get the translation set on the rendering context. It will be in
	// displayunits (i.e. pixels*scale), *not* pixels:
	nsTransform2D* xform;
	mMozContext->GetCurrentTransform(xform);
	float dx, dy;
	xform->GetTranslation(&dx, &dy);
	CGContextTranslateCTM(mCGContext, dx, dy);

// finish up	
	*_retval = mMozContext;
	NS_ADDREF(*_retval);
	return NS_OK;
}

NS_IMETHODIMP nsSVGQuartzCanvas::UnlockRenderingContext()
{
	OSErr err = ::QDEndCGContext(mGrafPtr, &mCGContext);
	NS_ENSURE_TRUE(err == noErr, NS_ERROR_FAILURE);
	return NS_OK;
}

NS_IMETHODIMP nsSVGQuartzCanvas::GetPresContext(nsIPresContext **_retval)
{
  *_retval = mPresContext;
  NS_IF_ADDREF(*_retval);
  return NS_OK;
}

NS_IMETHODIMP nsSVGQuartzCanvas::Clear(nscolor color)
{
	return NS_OK;
}

NS_IMETHODIMP nsSVGQuartzCanvas::Flush()
{
	CGContextFlush(mCGContext);
	return NS_OK;
}

//----------------------------------------------------------------------
// nsISVGQuartzCanvas methods:

NS_IMETHODIMP_(CGContextRef) nsSVGQuartzCanvas::GetCGContext()
{
	return mCGContext;
}

