/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Mozilla SVG project.
 *
 * The Initial Developer of the Original Code is 
 * Crocodile Clips Ltd..
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Alex Fritze <alex.fritze@crocodile-clips.com> (original author)
 *	  James Turner <james.turner@crocodile-clips.com> (Quartz port)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

#include "nsCOMPtr.h"
#include "nsSVGQuartzPath.h"
#include "nsISVGRendererPathGeometry.h"
#include "nsIDOMSVGMatrix.h"
#include "nsISVGRendererRegion.h"
#include "nsISVGPathGeometrySource.h"
#include "nsISVGRendererPathBuilder.h"
#include "nsISVGQuartzCanvas.h"
#include "nsMemory.h"


extern nsresult 
NS_NewSVGQuartzPathBuilder(nsISVGRendererPathBuilder **result,
	nsISVGPathGeometrySource *src, CGMutablePathRef output);


extern nsresult 
NS_NewSVGQuartzRegion(nsISVGRendererRegion** result, const RgnHandle rgn);

////////////////////////////////////////////////////////////////////////
// nsSVGQuartzPath class

class nsSVGQuartzPath : public nsISVGRendererPathGeometry
{
protected:
  friend nsresult NS_NewSVGQuartzPath(nsISVGRendererPathGeometry **result,
                                               nsISVGPathGeometrySource *src);

  nsSVGQuartzPath();
  virtual ~nsSVGQuartzPath();
  
  nsresult Init(nsISVGPathGeometrySource* src);
public:
  // nsISupports interface:
  NS_DECL_ISUPPORTS
  
  // nsISVGRendererPathGeometry interface:
  NS_DECL_NSISVGRENDERERPATHGEOMETRY
  
protected:
 
  CGPathRef GetPath();
  
  void SetupStrokeStyle(CGContextRef cg);
  void MapLineCap(CGContextRef cg, PRUint16 linecap);
  void MapLineJoin(CGContextRef cg, PRUint16 linejoin);
  
  void SetupTransform(CGContextRef cg);
 
 	RgnHandle GetStrokeRegion();
 	RgnHandle GetFillRegion();
 
private:
  nsCOMPtr<nsISVGPathGeometrySource> mSource;

  CGPathRef m_path;
  CGAffineTransform m_ctm;
  RgnHandle m_strokeRgn, m_fillRgn;
  bool m_pathValid, m_ctmValid;

//  nsCOMPtr<nsISVGRendererRegion> mCoveredRegion;
};

//----------------------------------------------------------------------
// implementation:

nsSVGQuartzPath::nsSVGQuartzPath() :
	m_strokeRgn(NULL), m_fillRgn(NULL),
	m_pathValid(false), m_ctmValid(false)
{
  NS_INIT_ISUPPORTS();
}

nsSVGQuartzPath::~nsSVGQuartzPath()
{
	if (m_pathValid)
		CGPathRelease(m_path);
		
	DisposeRgn(m_strokeRgn);
	DisposeRgn(m_fillRgn);
}

nsresult nsSVGQuartzPath::Init(nsISVGPathGeometrySource* src)
{
  mSource = src;
  return NS_OK;
}


nsresult
NS_NewSVGQuartzPath(nsISVGRendererPathGeometry **result,
                             nsISVGPathGeometrySource *src)
{
  nsSVGQuartzPath* path = new nsSVGQuartzPath();
  if (!path) return NS_ERROR_OUT_OF_MEMORY;

  NS_ADDREF(path);

  nsresult rv = path->Init(src);

  if (NS_FAILED(rv)) {
    NS_RELEASE(path);
    return rv;
  }
  
  *result = path;
  return rv;
}

//----------------------------------------------------------------------
// nsISupports methods:

NS_IMPL_ISUPPORTS1(nsSVGQuartzPath, nsISVGRendererPathGeometry)

//----------------------------------------------------------------------

CGPathRef
nsSVGQuartzPath::GetPath()
{
  if (m_pathValid) return  m_path;

  CGMutablePathRef mutablePath = CGPathCreateMutable();
  nsCOMPtr<nsISVGRendererPathBuilder> builder;
  NS_NewSVGQuartzPathBuilder(getter_AddRefs(builder), mSource, mutablePath);
  mSource->ConstructPath(builder);
  
  // copy to an immutable path here for speed?
  m_path = mutablePath;
  m_pathValid = true;
  return m_path;
}

RgnHandle nsSVGQuartzPath::GetStrokeRegion()
{
	if (m_strokeRgn && m_pathValid)
		return m_strokeRgn;
	::DisposeRgn(m_strokeRgn); // clean up the old one
	
	// compute width and height somehow ...
	Rect gworldBounds;
	CGRect bounds = CGPathGetBoundingBox(GetPath());
	
	 GWorldPtr mask = NULL;
	 OSErr err = NewGWorld(&mask, 1, &gworldBounds, NULL, NULL, 0);
	 NS_ASSERTION(err == noErr, "failed to create a 1-bit GWorld");
	 	
	CGContextRef ctx;
	 ::QDBeginCGContext(mask,&ctx);
	 
	 SetupStrokeStyle(ctx);
	 
	 CGContextBeginPath(ctx);
	 CGContextAddPath(ctx, GetPath());
	 
	 SetupTransform(ctx);
	 
	 CGContextStrokePath(ctx);
	 
	 ::QDEndCGContext(mask,&ctx);
	 
	 m_strokeRgn = NewRgn();
	 PixMapHandle pixMap = GetGWorldPixMap(mask);
	 BitMapToRegion(m_strokeRgn, (BitMap*) *pixMap);
	 
	 DisposeGWorld(mask);
	 
	 return m_strokeRgn;
}

RgnHandle nsSVGQuartzPath::GetFillRegion()
{
	if (m_fillRgn && m_pathValid)
		return m_fillRgn;
	::DisposeRgn(m_fillRgn); // clean up the old one
	
	// compute width and height somehow ...
	Rect gworldBounds;
	CGRect bounds = CGPathGetBoundingBox(GetPath());
	
	 GWorldPtr mask = NULL;
	 OSErr err = NewGWorld(&mask, 1, &gworldBounds, NULL, NULL, 0);
	 NS_ASSERTION(err == noErr, "failed to create a 1-bit GWorld");
	 	
	CGContextRef ctx;
	 ::QDBeginCGContext(mask,&ctx);
	 
	
	 
	 CGContextBeginPath(ctx);
	 CGContextAddPath(ctx, GetPath());
	 
	 SetupTransform(ctx);
	 
	 CGContextStrokePath(ctx);
	 
	 ::QDEndCGContext(mask,&ctx);
	 
	 m_fillRgn = NewRgn();
	 PixMapHandle pixMap = GetGWorldPixMap(mask);
	 BitMapToRegion(m_fillRgn, (BitMap*) *pixMap);
	 
	 DisposeGWorld(mask);
	 
	 return m_fillRgn;
}

void nsSVGQuartzPath::MapLineCap(CGContextRef cg, PRUint16 linecap)
{
	switch (linecap) {
	case nsISVGGeometrySource::STROKE_LINECAP_BUTT:		CGContextSetLineCap(cg, kCGLineCapButt);  break;
	case nsISVGGeometrySource::STROKE_LINECAP_ROUND:	CGContextSetLineCap(cg, kCGLineCapRound); break;
	case nsISVGGeometrySource::STROKE_LINECAP_SQUARE:	CGContextSetLineCap(cg, kCGLineCapSquare);break;
	default:
		NS_ASSERTION(PR_FALSE, "unknown line cap in SVG Quartz path");	
	}
}

void nsSVGQuartzPath::MapLineJoin(CGContextRef cg, PRUint16 linejoin)
{
	switch (linejoin) {
	case nsISVGGeometrySource::STROKE_LINEJOIN_MITER:	CGContextSetLineJoin(cg, kCGLineJoinMiter); break;
	case nsISVGGeometrySource::STROKE_LINEJOIN_ROUND:	CGContextSetLineJoin(cg, kCGLineJoinRound); break;
	case nsISVGGeometrySource::STROKE_LINEJOIN_BEVEL:	CGContextSetLineJoin(cg, kCGLineJoinBevel);	break;
	default:
		NS_ASSERTION(PR_FALSE, "unknown line join in SVG Quartz path");	
	}	
}

void nsSVGQuartzPath::SetupStrokeStyle(CGContextRef cg)
{
	nscolor color;
	float opacity;
	
	mSource->GetStrokePaint(&color);
    mSource->GetStrokeOpacity(&opacity);
	CGContextSetRGBStrokeColor(cg, NS_GET_R(color) / 256.0, 
		NS_GET_G(color) / 256.0, NS_GET_B(color) / 256.0, opacity);
	
	float width;
	mSource->GetStrokeWidth(&width);
	CGContextSetLineWidth(cg, width);
	
	PRUint16 lineCap, lineJoin;
	mSource->GetStrokeLinecap(&lineCap);
	MapLineCap(cg, lineCap);
	
	mSource->GetStrokeLinejoin(&lineJoin);
	MapLineJoin(cg, lineJoin);
	
	if (lineJoin == nsISVGGeometrySource::STROKE_LINEJOIN_MITER) {
		float miterLimit;
		mSource->GetStrokeMiterlimit(&miterLimit);
		CGContextSetMiterLimit(cg, miterLimit);	
	}
	
	// dashing
	float *dashArray;
 	PRUint32 count;
	mSource->GetStrokeDashArray(&dashArray, &count);
 
 	if (count>0) {
 		// scale dashes based on width
    	//for (PRUint32 i=0; i<count; ++i) dashArray[i]/=width;
    	float phase;
    	mSource->GetStrokeDashoffset(&phase);
    	
    	CGContextSetLineDash(cg, phase, dashArray, count);
   		nsMemory::Free(dashArray);
  	}	
}

void nsSVGQuartzPath::SetupTransform(CGContextRef cg)
{
	if (!m_ctmValid) {
		nsCOMPtr<nsIDOMSVGMatrix> ctm;
  		mSource->GetCTM(getter_AddRefs(ctm));
  
		ctm->GetA(&m_ctm.a);
		ctm->GetB(&m_ctm.b);
		ctm->GetC(&m_ctm.c);
		ctm->GetD(&m_ctm.d);
		
		ctm->GetE(&m_ctm.tx);
		ctm->GetF(&m_ctm.ty);
		
		m_ctmValid = true;	
	}
	
	CGContextConcatCTM(cg, m_ctm);
}

//----------------------------------------------------------------------
// nsISVGRendererPathGeometry methods:

NS_IMETHODIMP
nsSVGQuartzPath::Render(nsISVGRendererCanvas *svgCtx)
{
	nsCOMPtr<nsISVGQuartzCanvas> quartzCtx = do_QueryInterface(svgCtx);
	CGContextRef cgCtx = quartzCtx->GetCGContext();

	CGContextSaveGState(cgCtx);

  nscolor color;
  float opacity;
  PRUint16 fillType, strokeType;
  
  mSource->GetFillPaintType(&fillType);
  if (fillType == nsISVGGeometrySource::PAINT_TYPE_SOLID_COLOR) {
    mSource->GetFillPaint(&color);
    mSource->GetFillOpacity(&opacity);
    CGContextSetRGBFillColor(cgCtx, NS_GET_R(color) / 256.0, 
		NS_GET_G(color) / 256.0, NS_GET_B(color) / 256.0, opacity);
  }

  mSource->GetStrokePaintType(&strokeType);
  if (strokeType == nsISVGGeometrySource::PAINT_TYPE_SOLID_COLOR)
		SetupStrokeStyle(cgCtx);
  
  	SetupTransform(cgCtx);
  
	CGContextBeginPath(cgCtx);
	CGContextAddPath(cgCtx, GetPath());
	CGContextClosePath(cgCtx);
	
	CGPathDrawingMode drawMode;
	PRUint16  evenOdd;
	mSource->GetFillRule(&evenOdd);
	
	// horibble bit of logic to evalute various cases and compute the Quartz dra wing mode
	if (strokeType == nsISVGGeometrySource::PAINT_TYPE_SOLID_COLOR) {
		if (fillType == nsISVGGeometrySource::PAINT_TYPE_SOLID_COLOR)
			drawMode = (evenOdd == nsISVGGeometrySource::FILL_RULE_EVENODD) ? kCGPathEOFillStroke : kCGPathFillStroke;
		else
			drawMode = kCGPathStroke;
	} else if (fillType == nsISVGGeometrySource::PAINT_TYPE_SOLID_COLOR)
		drawMode = (evenOdd == nsISVGGeometrySource::FILL_RULE_EVENODD) ? kCGPathEOFill : kCGPathFill;

	CGContextDrawPath(cgCtx, drawMode);	
	CGContextRestoreGState(cgCtx);
	
  return NS_OK;
}

/* nsISVGRendererRegion update (in unsigned long updatemask); */
NS_IMETHODIMP
nsSVGQuartzPath::Update(PRUint32 updatemask, nsISVGRendererRegion **_retval)
{
  *_retval = nsnull;

  const unsigned long pathmask =
    nsISVGGeometrySource::UPDATEMASK_FILL_RULE |
    nsISVGPathGeometrySource::UPDATEMASK_PATH;

  const unsigned long fillmask = 
    pathmask |
    nsISVGGeometrySource::UPDATEMASK_CTM;

  const unsigned long strokemask =
    pathmask |
    nsISVGGeometrySource::UPDATEMASK_STROKE_WIDTH       |
    nsISVGGeometrySource::UPDATEMASK_STROKE_LINECAP     |
    nsISVGGeometrySource::UPDATEMASK_STROKE_LINEJOIN    |
    nsISVGGeometrySource::UPDATEMASK_STROKE_MITERLIMIT  |
    nsISVGGeometrySource::UPDATEMASK_STROKE_DASH_ARRAY  |
    nsISVGGeometrySource::UPDATEMASK_STROKE_DASHOFFSET  |
    nsISVGGeometrySource::UPDATEMASK_CTM;

  const unsigned long hittestregionmask =
    fillmask                                            |
    strokemask                                          |
    nsISVGPathGeometrySource::UPDATEMASK_HITTEST_MASK;

  const unsigned long coveredregionmask =
    fillmask                                            |
    strokemask                                          |
    nsISVGGeometrySource::UPDATEMASK_FILL_PAINT_TYPE    |
    nsISVGGeometrySource::UPDATEMASK_STROKE_PAINT_TYPE;
  
  nsCOMPtr<nsISVGRendererRegion> before;
  GetCoveredRegion(getter_AddRefs(before));

  if (updatemask & pathmask) m_pathValid = false;
  if (updatemask & fillmask) {
  	DisposeRgn(m_fillRgn);
  	m_fillRgn = NULL;
  }
  
  if (updatemask & strokemask) {
  	DisposeRgn(m_strokeRgn);
  	m_strokeRgn = NULL;
  }
  
  if (updatemask & coveredregionmask) {
    nsCOMPtr<nsISVGRendererRegion> after;
    GetCoveredRegion(getter_AddRefs(after));
    if (after)
      after->Combine(before, _retval);
  } else if (updatemask != nsISVGGeometrySource::UPDATEMASK_NOTHING) {
    *_retval = before;
    NS_IF_ADDREF(*_retval);
  }
  return NS_OK;
}

/* nsISVGRendererRegion getCoveredRegion (); */
NS_IMETHODIMP
nsSVGQuartzPath::GetCoveredRegion(nsISVGRendererRegion **_retval)
{
  *_retval = nsnull;
  
  RgnHandle covered = NewRgn();
  UnionRgn(GetStrokeRegion(), GetFillRegion(), covered);
  
  return NS_NewSVGQuartzRegion(_retval, covered);
}

/* boolean containsPoint (in float x, in float y); */
NS_IMETHODIMP
nsSVGQuartzPath::ContainsPoint(float x, float y, PRBool *_retval)
{
  *_retval = PR_FALSE;
  Point pt = {static_cast<short>(x), static_cast<short>(y)};
  
  PRUint16 hitMask=0;
  mSource->GetHittestMask(&hitMask);
  
  if ((hitMask & nsISVGPathGeometrySource::HITTEST_MASK_FILL) && 
  	PtInRgn(pt, GetFillRegion()))
  	*_retval = PR_TRUE;	
    
  if ((hitMask & nsISVGPathGeometrySource::HITTEST_MASK_STROKE) && 
  	PtInRgn(pt, GetStrokeRegion()))
    *_retval = PR_TRUE;

  return NS_OK;
}
