/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Mozilla SVG project.
 *
 * The Initial Developer of the Original Code is 
 * Crocodile Clips Ltd..
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Alex Fritze <alex.fritze@crocodile-clips.com> (original author)
 *    James Turner <james.turner@crocodile-clips.com> (Quartz port)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or 
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

#include "nsCOMPtr.h"
#include "nsISVGRectangleSink.h"
#include "nsIPresContext.h"
#include "nsISVGQuartzRegion.h"

////////////////////////////////////////////////////////////////////////
// nsSVGQuartzRegion class

class nsSVGQuartzRegion : public nsISVGQuartzRegion
{
public:
  nsSVGQuartzRegion(const RgnHandle rgn);
  virtual ~nsSVGQuartzRegion();

  // nsISupports interface:
  NS_DECL_ISUPPORTS

  // nsISVGRendererRegion interface:
  NS_DECL_NSISVGRENDERERREGION
  
  // nsISVGQuartzRegion interface
  NS_IMETHOD_(const RgnHandle) GetRegion() const;
private:
	RgnHandle mRegion;
};

//----------------------------------------------------------------------
// implementation:


nsSVGQuartzRegion::nsSVGQuartzRegion(const RgnHandle rgn) :
	mRegion(rgn)
{
  NS_INIT_ISUPPORTS();
}

nsSVGQuartzRegion::~nsSVGQuartzRegion()
{
	if (mRegion)
		DisposeRgn(mRegion);
}

nsresult
NS_NewSVGQuartzRegion(nsISVGRendererRegion** result, const RgnHandle rgn)
{
  *result = new nsSVGQuartzRegion(rgn);
  if (!*result) return NS_ERROR_OUT_OF_MEMORY;
  NS_ADDREF(*result);
  return NS_OK;
}


nsresult
NS_NewSVGQuartzRegion(nsISVGRendererRegion** result, float x, float y, float w, float h)
{
	RgnHandle rr = NewRgn();
	Rect qdRect;
	SetRect(&qdRect, (short)round(x), (short)round(y), (short)round(x+w), (short)round(y+h));
	RectRgn(rr, &qdRect);
	
  *result = new nsSVGQuartzRegion(rr);
  if (!*result) return NS_ERROR_OUT_OF_MEMORY;
  NS_ADDREF(*result);
  return NS_OK;
}

//----------------------------------------------------------------------
// nsISupports methods:

NS_IMPL_ADDREF(nsSVGQuartzRegion)
NS_IMPL_RELEASE(nsSVGQuartzRegion)

NS_INTERFACE_MAP_BEGIN(nsSVGQuartzRegion)
  NS_INTERFACE_MAP_ENTRY(nsISVGQuartzRegion)
  NS_INTERFACE_MAP_ENTRY(nsISVGRendererRegion)
  NS_INTERFACE_MAP_ENTRY(nsISupports)
NS_INTERFACE_MAP_END

NS_IMETHODIMP_(const RgnHandle)
nsSVGQuartzRegion::GetRegion() const
{
  return mRegion;
}

//----------------------------------------------------------------------
// nsISVGRendererRegion methods:

/* nsISVGRendererRegion combine (in nsISVGRendererRegion other); */
NS_IMETHODIMP
nsSVGQuartzRegion::Combine(nsISVGRendererRegion *other, nsISVGRendererRegion **_retval)
{
  *_retval = nsnull;

  nsSVGQuartzRegion *union_region = new nsSVGQuartzRegion(::NewRgn());
  NS_ADDREF(union_region);
  *_retval = union_region;

  if (!other) {
  	::CopyRgn(mRegion, union_region->GetRegion());
  	return NS_OK; // we're all done
  }
  
  nsCOMPtr<nsISVGQuartzRegion> quartzOther = do_QueryInterface(other);
  NS_ENSURE_TRUE(quartzOther, NS_ERROR_NO_INTERFACE);

  ::UnionRgn(mRegion, quartzOther->GetRegion(), union_region->GetRegion());
  return NS_OK;
}

OSStatus RegionToRectsCallback (UInt16 message, RgnHandle rgn, const Rect *rect, void *refCon)
{
	if (message != kQDRegionToRectsMsgParse) return noErr;
	nsISVGRectangleSink *sink = static_cast<nsISVGRectangleSink*>(refCon);
	sink->SinkRectangle(rect->left, rect->top, 
		rect->right - rect->left, rect->bottom - rect->top);
	return noErr;
}

/* void getRectangleScans (in nsISVGRectangleSink sink); */
NS_IMETHODIMP
nsSVGQuartzRegion::GetRectangleScans(nsISVGRectangleSink *sink)
{
	NS_ENSURE_ARG_POINTER(sink);
	OSErr err = ::QDRegionToRects(mRegion, kQDParseRegionFromTop, &RegionToRectsCallback, sink);
	NS_ENSURE_TRUE(err == noErr, NS_ERROR_FAILURE);
 	return NS_OK;
}
